import os
import sys
import tempfile
import zipfile
from sys import platform
import argparse
import requests
import pyrfuniverse
from pyrfuniverse.utils.version import Version


def pyrfuniverse_entry_points():
    parser = argparse.ArgumentParser(description='RFUniverse entry points')
    subparsers = parser.add_subparsers(dest='command', help='download/etc')
    parser_download = subparsers.add_parser('download', help='Download rfuniverse release')
    parser_download.add_argument('-s', '--savepath', type=str, help='RFUniverse release save path')
    parser_download.add_argument('-v', '--version', type=str, help='RFUniverse release Version')
    args = parser.parse_args()
    if args.command == 'download':
        download(args)


def download(args):
    if args.savepath is None:
        save_path = os.path.expanduser(r"~/rfuniverse")
    else:
        save_path = args.savepath

    releases_url = f"https://api.github.com/repos/robotflow-initiative/rfuniverse/releases"
    response = requests.get(releases_url)
    releases = response.json()
    last_version = None

    if args.version is None:
        current_version = Version(pyrfuniverse.__version__)
        for i in releases:
            if "v" not in i["tag_name"] and "." not in i["tag_name"]:
                continue
            version = Version(i["tag_name"].replace("v", ""))
            if version[0] == current_version[0] and version[1] == current_version[1] and version[2] == current_version[2]:
                if last_version is None:
                    last_version = version
                elif version[3] > current_version[3]:
                    last_version = version
    else:
        current_version = Version(args.version)
        for i in releases:
            if "v" not in i["tag_name"] and "." not in i["tag_name"]:
                continue
            version = Version(i["tag_name"].replace("v", ""))
            if version == current_version:
                last_version = version

    if last_version is None:
        print("No available version found")
        return

    if platform == "win32":
        platform_string = "Windows"
        suffix = ".exe"
    elif platform == "linux":
        platform_string = "Linux"
        suffix = ".x86_64"
    else:
        raise EnvironmentError(f"Unsupported systems: {platform}")

    download_url = None
    for i in releases:
        if i["tag_name"] == f"v{last_version}":
            for asset in i["assets"]:
                if platform_string in asset["name"]:
                    download_url = asset["browser_download_url"]
                    download_size = asset["size"]
                    break
            break

    if download_url is None:
        print("No available version found")
        return

    temp_path = os.path.join(tempfile.gettempdir(), "temp_rfu_release.zip")
    with requests.get(download_url, stream=True) as response:
        print(f"Downloading: {download_url}")
        print(f"---")
        length = 0
        with open(temp_path, 'wb') as f:
            for chunk in response.iter_content(chunk_size=1024000):
                length = length + 1024000
                f.write(chunk)
                sys.stdout.write("\033[F")
                sys.stdout.write("\033[K")
                print(f"{length}/{download_size}---{int(length * 100 / download_size)}%")

    os.makedirs(save_path, exist_ok=True)
    with zipfile.ZipFile(temp_path, 'r') as zipObject:
        zipObject.extractall(save_path)
    folder_name = os.path.splitext(os.path.basename(download_url))[0]
    folder_path = os.path.join(save_path, folder_name)
    print(fr"Unzip Done: {folder_path}")
    executable_file = os.path.abspath(os.path.join(folder_path, f"RFUniverse{suffix}"))
    pyrfuniverse.config["executable_file"] = executable_file
    pyrfuniverse.save_config(pyrfuniverse.config)
    if platform == "linux":
        os.system(f"chmod +x {executable_file}")
    print(fr"Write config executable_file: {executable_file}")



import pyrfuniverse.attributes as attr


class RigidbodyAttr(attr.ColliderAttr):
    """
    Rigid body class.
    """

    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.

            self.data['velocity']: The velocity of the object.

            self.data['angular_velocity']: The angular velcity of the object.
        """
        super().parse_message(data)

    def SetMass(self, mass: float):
        """
        Set the mass of this rigid body object

        Args:
            mass: Float, representing the mass of this rigid body.
        """
        self._send_data("SetMass", float(mass))

    def SetDrag(self, drag: float):
        """
        Set the drag of this rigid body object

        Args:
            drag: Float, representing the drag of this rigid body.
        """
        self._send_data("SetDrag", float(drag))

    def SetAngularDrag(self, angular_drag: float):
        """
        Set the angular drag of this rigid body object

        Args:
            angular_drag: Float, representing the angular drag of this rigid body.
        """
        self._send_data("SetAngularDrag", float(angular_drag))
    def SetUseGravity(self, use_gravity: bool):
        """
        Set the rigid body use gravity or not.

        Args:
            use_gravity: Bool, use gravity or not.
        """
        self._send_data("SetUseGravity", use_gravity)

    def EnabledMouseDrag(self, enabled: bool):
        """
        Enable or Disable the rigid body Mouse Drag.

        Args:
            enabled: Bool, Enabled Mouse Drag or not.
        """
        self._send_data("EnabledMouseDrag", enabled)

    def AddForce(self, force: list):
        """
        Add force to this rigid body object.

        Args:
            force: A list of length 3, representing the force added to this rigid body.
        """
        if force is not None:
            force = [float(i) for i in force]

        self._send_data("AddForce", force)

    def SetVelocity(self, velocity: list):
        """
        Set the velocity of this rigid body object.

        Args:
            velocity: A list of length 3, representing the velocity of this rigid body.
        """
        if velocity is not None:
            velocity = [float(i) for i in velocity]

        self._send_data("SetVelocity", velocity)

    def SetAngularVelocity(self, angular_velocity: list):
        """
        Set the angular velocity of this rigid body object.

        Args:
            angular_velocity: A list of length 3, representing the angular velocity of this rigid body.
        """
        if angular_velocity is not None:
            angular_velocity = [float(i) for i in angular_velocity]

        self._send_data("SetAngularVelocity", angular_velocity)

    def SetKinematic(self, is_kinematic: bool):
        """
        Set the Rigidbody is kinematic or not.

        Args:
            is_kinematic: is kinematic or not.
        """
        self._send_data("SetKinematic", is_kinematic)

    def Link(self, target_id: int, joint_index: int = 0, mass_scale: float = 1, connected_mass_scale: float = 1):
        """
        Link this rigidbody to another rigidbody or ArticulationBody

        Args:
            target_id: id of another rigidbody or ControllerAttr.
            joint_index: id of ControllerAttr joint.
            mass_scale: The scale to apply to the inverse mass and inertia tensor of the body prior to solving the constraints.
            connected_mass_scale: The scale to apply to the inverse mass and inertia tensor of the connected body prior to solving the constraints.
        """
        self._send_data("Link", int(target_id), int(joint_index), float(mass_scale), float(connected_mass_scale))



import pyrfuniverse.attributes as attr


class ClothAttr(attr.BaseAttr):
    """
    Obi cloth class.
    """
    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.

        """
        super().parse_message(data)

    def GetParticles(self):
        """
        get the cloth particles.
        """
        self._send_data("GetParticles")

    def AddAttach(self, id: int, max_dis: float = 0.03):
        """
        Add Attach clothing to attr object with given ID

        Args:
            id: Int, Tatget attr object id.
            max_dis: Float, max distance.
        """
        self._send_data("AddAttach", int(id), float(max_dis))

    def RemoveAttach(self, id: int):
        """
        Remove Attach clothing to attr object with given ID

        Args:
            id: Int, Tatget attr object id.
        """
        self._send_data("RemoveAttach", int(id))



class Version():
    def __init__(self, version_string: str):
        self.raw_string = version_string
        self.version = self.raw_string.split(".")
        self.version = [int(i) for i in self.version]
        while len(self.version) < 4:
            self.version.append(0)

    def __lt__(self, other):
        for i in range(4):
            if self.version[i] < other.version[i]:
                return True
            elif self.version[i] == other.version[i]:
                continue
        return False

    def __gt__(self, other):
        for i in range(4):
            if self.version[i] > other.version[i]:
                return True
            elif self.version[i] == other.version[i]:
                continue
        return False

    def __eq__(self, other):
        for i in range(4):
            if self.version[i] != other.version[i]:
                return False
        return True

    def __str__(self):
        return self.raw_string

    def __repr__(self):
        return self.raw_string

    def toString(self):
        return self.raw_string

    def __getitem__(self, index):
        return self.version[index]

    def __setitem__(self, index, value):
        self.version[index] = value

    @staticmethod
    def sorted(versions: list, reverse=False):
        return sorted(versions, key=lambda v: v.version, reverse=reverse)






import numpy as np


class CoordinateSystemConverter():
    """
    Coordinate System Converter class.

    Args:
        cs1_direction: list, The visual direction corresponding to the xyz axis,
                        can be:
                        ["left"/"right"/"up"/"down"/"forward"/"back"/
                        "l"/"r"/"u"/"d"/"f"/"b"/
                        "-left"/"-right"/"-up"/"-down"/"-forward"/"-back"/
                        "-l"/"-r"/"-u"/"-d"/"-f"/"-b"]
        cs2_direction: list, The visual direction corresponding to the xyz axis,
                        can be:
                        ["left"/"right"/"up"/"down"/"forward"/"back"/
                        "l"/"r"/"u"/"d"/"f"/"b"/
                        "-left"/"-right"/"-up"/"-down"/"-forward"/"-back"/
                        "-l"/"-r"/"-u"/"-d"/"-f"/"-b"]
    """
    def __init__(self, cs1_direction=["right", "up", "forward"], cs2_direction=["right", "up", "forward"]):
        self.cs1_dir, self.cs1_axis, self.cs1_sign, self.cs1_system = self._preprocessing(cs1_direction)
        self.cs2_dir, self.cs2_axis, self.cs2_sign, self.cs2_system = self._preprocessing(cs2_direction)

        self.cs1_to_cs2_map = [
            self.cs1_axis[self.cs2_dir[0]],
            self.cs1_axis[self.cs2_dir[1]],
            self.cs1_axis[self.cs2_dir[2]],
        ]
        self.cs1_to_cs2_sign = [
            self.cs1_sign[self.cs2_dir[0]] * self.cs2_sign[self.cs2_dir[0]],
            self.cs1_sign[self.cs2_dir[1]] * self.cs2_sign[self.cs2_dir[1]],
            self.cs1_sign[self.cs2_dir[2]] * self.cs2_sign[self.cs2_dir[2]],
        ]
        self.cs2_to_cs1_map = [
            self.cs2_axis[self.cs1_dir[0]],
            self.cs2_axis[self.cs1_dir[1]],
            self.cs2_axis[self.cs1_dir[2]],
        ]
        self.cs2_to_cs1_sign = [
            self.cs1_sign[self.cs1_dir[0]] * self.cs2_sign[self.cs1_dir[0]],
            self.cs1_sign[self.cs1_dir[1]] * self.cs2_sign[self.cs1_dir[1]],
            self.cs1_sign[self.cs1_dir[2]] * self.cs2_sign[self.cs1_dir[2]],
        ]
        self.cs1_left_or_right = self.cs1_sign['right'] * self.cs1_sign['up'] * self.cs1_sign['forward']
        self.cs2_left_or_right = self.cs2_sign['right'] * self.cs2_sign['up'] * self.cs2_sign['forward']

    def _preprocessing(self, direction):
        dir = ["", "", ""]
        axis = {}
        sign = {}
        direction_in_left = [[], [], []]
        for i in range(3):
            if direction[i] in ["right", "r", "-l", "-left"]:
                dir[i] = "right"
                axis["right"] = i
                sign["right"] = 1
                direction_in_left[i] = [1, 0, 0]
            elif direction[i] in ["up",  "u",  "-d",  "-down"]:
                dir[i] = "up"
                axis["up"] = i
                sign["up"] = 1
                direction_in_left[i] = [0, 1, 0]
            elif direction[i] in ["forward",  "f",  "-b",  "-back"]:
                dir[i] = "forward"
                axis["forward"] = i
                sign["forward"] = 1
                direction_in_left[i] = [0, 0, 1]
            elif direction[i] in ["-right",  "-r",  "l",  "left"]:
                dir[i] = "right"
                axis["right"] = i
                sign["right"] = -1
                direction_in_left[i] = [-1, 0, 0]
            elif direction[i] in ["-up",  "-u",  "d",  "down"]:
                dir[i] = "up"
                axis["up"] = i
                sign["up"] = -1
                direction_in_left[i] = [0, -1, 0]
            elif direction[i] in ["-forward",  "-f",  "b",  "back"]:
                dir[i] = "forward"
                axis["forward"] = i
                sign["forward"] = -1
                direction_in_left[i] = [0, 0, -1]
            else:
                raise f"{direction[i]} is Unrecognized axis"

        z = self._cross(direction_in_left[0], direction_in_left[1])
        if z == direction_in_left[2]:
            system = 1
        else:
            system = -1
        return dir, axis, sign, system

    def _cross(self, lhs: list, rhs: list) -> list:
        return [lhs[1] * rhs[2] - lhs[2] * rhs[1], lhs[2] * rhs[0] - lhs[0] * rhs[2], lhs[0] * rhs[1] - lhs[1] * rhs[0]]

    def cs1_pos_to_cs2_pos(self, pos: list) -> list:
        """
        Convert position form Coordinate System 1 to Coordinate System 2.

        Args:
            pos: List of length 3, position of Coordinate System 1.

        Return:
            list: List of length 3, position of Coordinate System 2.
        """
        x = pos[self.cs1_to_cs2_map[0]] * self.cs1_to_cs2_sign[0]
        y = pos[self.cs1_to_cs2_map[1]] * self.cs1_to_cs2_sign[1]
        z = pos[self.cs1_to_cs2_map[2]] * self.cs1_to_cs2_sign[2]
        return [x, y, z]

    def cs2_pos_to_cs1_pos(self, pos: list) -> list:
        """
        Convert position form Coordinate System 2 to Coordinate System 1.

        Args:
            pos: List of length 3, position of Coordinate System 2.

        Return:
            list: List of length 3, position of Coordinate System 1.
        """
        x = pos[self.cs2_to_cs1_map[0]] * self.cs2_to_cs1_sign[0]
        y = pos[self.cs2_to_cs1_map[1]] * self.cs2_to_cs1_sign[1]
        z = pos[self.cs2_to_cs1_map[2]] * self.cs2_to_cs1_sign[2]
        return [x, y, z]

    def cs1_quat_to_cs2_quat(self, quat: list) -> list:
        """
        Convert quaternion form Coordinate System 1 to Coordinate System 2.

        Args:
            quat: List of length 3, quaternion of Coordinate System 1.

        Return:
            list: List of length 3, quaternion of Coordinate System 2.
        """
        x = quat[self.cs1_to_cs2_map[0]] * self.cs1_to_cs2_sign[0]
        y = quat[self.cs1_to_cs2_map[1]] * self.cs1_to_cs2_sign[1]
        z = quat[self.cs1_to_cs2_map[2]] * self.cs1_to_cs2_sign[2]
        w = quat[3] * self.cs1_system * self.cs2_system
        return [x, y, z, w]

    def cs2_quat_to_cs1_quat(self, quat: list) -> list:
        """
        Convert quaternion form Coordinate System 2 to Coordinate System 1.

        Args:
            quat: List of length 4, quaternion[x,y,z,w] of Coordinate System 2.

        Return:
            list: List of length 4, quaternion[x,y,z,w] of Coordinate System 1.
        """
        x = quat[self.cs2_to_cs1_map[0]] * self.cs2_to_cs1_sign[0]
        y = quat[self.cs2_to_cs1_map[1]] * self.cs2_to_cs1_sign[1]
        z = quat[self.cs2_to_cs1_map[2]] * self.cs2_to_cs1_sign[2]
        w = quat[3] * self.cs1_system * self.cs2_system
        return [x, y, z, w]

    def cs1_scale_to_cs2_scale(self, scale: list) -> list:
        """
        Convert scale form Coordinate System 1 to Coordinate System 2.

        Args:
            scale: List of length 3, scale of Coordinate System 1.

        Return:
            list: List of length 3, scale of Coordinate System 2.
        """
        x = scale[self.cs1_to_cs2_map[0]]
        y = scale[self.cs1_to_cs2_map[1]]
        z = scale[self.cs1_to_cs2_map[2]]
        return [x, y, z]

    def cs2_scale_to_cs1_scale(self, scale: list) -> list:
        """
        Convert scale form Coordinate System 2 to Coordinate System 1.

        Args:
            scale: List of length 3, scale of Coordinate System 2.

        Return:
            list: List of length 3, scale of Coordinate System 1.
        """
        x = scale[self.cs2_to_cs1_map[0]]
        y = scale[self.cs2_to_cs1_map[1]]
        z = scale[self.cs2_to_cs1_map[2]]
        return [x, y, z]

    def cs1_matrix_to_cs2_matrix(self, matrix) -> np.ndarray:
        """
        Convert rotation matrix form Coordinate System 1 to Coordinate System 2.

        Args:
            matrix: list or np.ndarray shape[3,3], rotation matrix of Coordinate System 1.

        Return:
            np.ndarray: shape[3,3], rotation matrix of Coordinate System 2.
        """
        quaternion = self.matrix_to_quat(matrix)
        quaternion = self.cs1_quat_to_cs2_quat(quaternion)
        return self.quat_to_matrix(quaternion)

    def cs2_matrix_to_cs1_matrix(self, matrix) -> np.ndarray:
        """
        Convert rotation matrix form Coordinate System 2 to Coordinate System 1.

        Args:
            matrix: list or np.ndarray shape[3,3], rotation matrix of Coordinate System 2.

        Return:
            np.ndarray: shape[3,3], rotation matrix of Coordinate System 1.
        """
        quaternion = self.matrix_to_quat(matrix)
        quaternion = self.cs2_quat_to_cs1_quat(quaternion)
        return self.quat_to_matrix(quaternion)

    def quat_to_matrix(self, quat=[0,0,0,1]) -> np.ndarray:
        return np.array([
            [1 - 2 * (quat[1] ** 2 + quat[2] ** 2), 2 * (quat[0] * quat[1] - quat[2] * quat[3]), 2 * (quat[0] * quat[2] + quat[1] * quat[3])],
            [2 * (quat[0] * quat[1] + quat[2] * quat[3]), 1 - 2 * (quat[0] ** 2 + quat[2] ** 2), 2 * (quat[1] * quat[2] - quat[0] * quat[3])],
            [2 * (quat[0] * quat[2] - quat[1] * quat[3]), 2 * (quat[1] * quat[2] + quat[0] * quat[3]), 1 - 2 * (quat[0] ** 2 + quat[1] ** 2)]
        ])

    def matrix_to_quat(self, matrix) -> list:
        m = np.array(matrix, dtype=float)
        trace = np.trace(m)
        if trace > 0:
            s = 0.5 / np.sqrt(trace + 1.0)
            w = 0.25 / s
            x = (m[2, 1] - m[1, 2]) * s
            y = (m[0, 2] - m[2, 0]) * s
            z = (m[1, 0] - m[0, 1]) * s
        elif m[0, 0] > m[1, 1] and m[0, 0] > m[2, 2]:
            s = 2.0 * np.sqrt(1.0 + m[0, 0] - m[1, 1] - m[2, 2])
            w = (m[2, 1] - m[1, 2]) / s
            x = 0.25 * s
            y = (m[0, 1] + m[1, 0]) / s
            z = (m[0, 2] + m[2, 0]) / s
        elif m[1, 1] > m[2, 2]:
            s = 2.0 * np.sqrt(1.0 + m[1, 1] - m[0, 0] - m[2, 2])
            w = (m[0, 2] - m[2, 0]) / s
            x = (m[0, 1] + m[1, 0]) / s
            y = 0.25 * s
            z = (m[1, 2] + m[2, 1]) / s
        else:
            s = 2.0 * np.sqrt(1.0 + m[2, 2] - m[0, 0] - m[1, 1])
            w = (m[1, 0] - m[0, 1]) / s
            x = (m[0, 2] + m[2, 0]) / s
            y = (m[1, 2] + m[2, 1]) / s
            z = 0.25 * s
        return [x, y, z, w]